# SPDX-FileCopyrightText: 2021-present Ofek Lev <oss@ofek.dev>
#
# SPDX-License-Identifier: MIT
import subprocess
import sys


def build_project(*args):
    process = subprocess.run(
        [sys.executable, '-m', 'hatchling', 'build', *args], stdout=subprocess.PIPE, stderr=subprocess.STDOUT
    )
    if process.returncode:  # no cov
        raise Exception(process.stdout.decode('utf-8'))



# SPDX-FileCopyrightText: 2021-present Ofek Lev <oss@ofek.dev>
#
# SPDX-License-Identifier: MIT
import os
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Generator

import pytest


@pytest.fixture
def temp_dir() -> Generator[Path, None, None]:
    with TemporaryDirectory() as d:
        yield Path(os.path.realpath(d))


@pytest.fixture
def new_project(temp_dir):
    project_dir = temp_dir / 'my-app'
    project_dir.mkdir()

    gitignore_file = project_dir / '.gitignore'
    gitignore_file.write_text('*.so\n*.pyd', encoding='utf-8')

    project_file = project_dir / 'pyproject.toml'
    project_file.write_text(
        """\
[build-system]
requires = ["hatchling"]
build-backend = "hatchling.build"

[project]
name = "my-app"
dynamic = ["version"]

[tool.hatch.version]
path = "my_app/__init__.py"

[tool.hatch.build]
packages = ["my_app"]

[tool.hatch.build.targets.wheel.hooks.mypyc]
dependencies = ["hatch-mypyc"]
include = ["/my_app"]
""",
        encoding='utf-8',
    )

    package_dir = project_dir / 'my_app'
    package_dir.mkdir()

    package_root = package_dir / '__init__.py'
    package_root.write_text('__version__ = "1.2.3"', encoding='utf-8')

    fibonacci_file = package_dir / 'fib.py'
    fibonacci_file.write_text(
        """\
def fib(n: int) -> int:
    if n <= 1:
        return n
    else:
        return fib(n - 2) + fib(n - 1)
""",
        encoding='utf-8',
    )

    package_data_file = package_dir / 'driver.yaml'
    package_data_file.write_text('apiVersion: storage.k8s.io/v1', encoding='utf-8')

    origin = os.getcwd()
    os.chdir(project_dir)
    try:
        yield project_dir
    finally:
        os.chdir(origin)



# SPDX-FileCopyrightText: 2021-present Ofek Lev <oss@ofek.dev>
#
# SPDX-License-Identifier: MIT
import pytest

from hatch_mypyc.plugin import MypycBuildHook


class TestMypyArgs:
    def test_correct(self, new_project):
        config = {'mypy-args': ['foo']}
        build_dir = new_project / 'dist'
        build_hook = MypycBuildHook(str(new_project), config, str(build_dir), 'wheel')

        assert build_hook.config_mypy_args == ['foo']

    def test_not_array(self, new_project):
        config = {'mypy-args': 9000}
        build_dir = new_project / 'dist'
        build_hook = MypycBuildHook(str(new_project), config, str(build_dir), 'wheel')

        with pytest.raises(TypeError, match='Option `mypy-args` for build hook `mypyc` must be an array'):
            _ = build_hook.config_mypy_args

    def test_argument_not_string(self, new_project):
        config = {'mypy-args': [9000]}
        build_dir = new_project / 'dist'
        build_hook = MypycBuildHook(str(new_project), config, str(build_dir), 'wheel')

        with pytest.raises(
            TypeError, match='Argument #1 of option `mypy-args` for build hook `mypyc` must be a string'
        ):
            _ = build_hook.config_mypy_args

    def test_argument_empty_string(self, new_project):
        config = {'mypy-args': ['']}
        build_dir = new_project / 'dist'
        build_hook = MypycBuildHook(str(new_project), config, str(build_dir), 'wheel')

        with pytest.raises(
            ValueError, match='Argument #1 of option `mypy-args` for build hook `mypyc` cannot be an empty string'
        ):
            _ = build_hook.config_mypy_args


class TestOptions:
    def test_correct(self, new_project):
        config = {'options': {'opt_level': '3'}}
        build_dir = new_project / 'dist'
        build_hook = MypycBuildHook(str(new_project), config, str(build_dir), 'wheel')

        assert build_hook.config_options == {'opt_level': '3'}

    def test_not_table(self, new_project):
        config = {'options': 9000}
        build_dir = new_project / 'dist'
        build_hook = MypycBuildHook(str(new_project), config, str(build_dir), 'wheel')

        with pytest.raises(TypeError, match='Option `options` for build hook `mypyc` must be a table'):
            _ = build_hook.config_options


class TestInclude:
    def test_correct(self, new_project):
        config = {'include': ['foo']}
        build_dir = new_project / 'dist'
        build_hook = MypycBuildHook(str(new_project), config, str(build_dir), 'wheel')

        assert build_hook.config_include == ['foo']

    def test_not_array(self, new_project):
        config = {'include': 9000}
        build_dir = new_project / 'dist'
        build_hook = MypycBuildHook(str(new_project), config, str(build_dir), 'wheel')

        with pytest.raises(TypeError, match='Option `include` for build hook `mypyc` must be an array'):
            _ = build_hook.config_include

    def test_missing(self, new_project):
        config = {'include': []}
        build_dir = new_project / 'dist'
        build_hook = MypycBuildHook(str(new_project), config, str(build_dir), 'wheel')

        with pytest.raises(ValueError, match='Option `include` for build hook `mypyc` is required'):
            _ = build_hook.config_include

    def test_pattern_not_string(self, new_project):
        config = {'include': [9000]}
        build_dir = new_project / 'dist'
        build_hook = MypycBuildHook(str(new_project), config, str(build_dir), 'wheel')

        with pytest.raises(TypeError, match='Pattern #1 of option `include` for build hook `mypyc` must be a string'):
            _ = build_hook.config_include

    def test_pattern_empty_string(self, new_project):
        config = {'include': ['']}
        build_dir = new_project / 'dist'
        build_hook = MypycBuildHook(str(new_project), config, str(build_dir), 'wheel')

        with pytest.raises(
            ValueError, match='Pattern #1 of option `include` for build hook `mypyc` cannot be an empty string'
        ):
            _ = build_hook.config_include


class TestExclude:
    def test_correct(self, new_project):
        config = {'exclude': ['foo']}
        build_dir = new_project / 'dist'
        build_hook = MypycBuildHook(str(new_project), config, str(build_dir), 'wheel')

        assert build_hook.config_exclude == ['foo']

    def test_not_array(self, new_project):
        config = {'exclude': 9000}
        build_dir = new_project / 'dist'
        build_hook = MypycBuildHook(str(new_project), config, str(build_dir), 'wheel')

        with pytest.raises(TypeError, match='Option `exclude` for build hook `mypyc` must be an array'):
            _ = build_hook.config_exclude

    def test_pattern_not_string(self, new_project):
        config = {'exclude': [9000]}
        build_dir = new_project / 'dist'
        build_hook = MypycBuildHook(str(new_project), config, str(build_dir), 'wheel')

        with pytest.raises(TypeError, match='Pattern #1 of option `exclude` for build hook `mypyc` must be a string'):
            _ = build_hook.config_exclude

    def test_pattern_empty_string(self, new_project):
        config = {'exclude': ['']}
        build_dir = new_project / 'dist'
        build_hook = MypycBuildHook(str(new_project), config, str(build_dir), 'wheel')

        with pytest.raises(
            ValueError, match='Pattern #1 of option `exclude` for build hook `mypyc` cannot be an empty string'
        ):
            _ = build_hook.config_exclude



# SPDX-FileCopyrightText: 2021-present Ofek Lev <oss@ofek.dev>
#
# SPDX-License-Identifier: MIT
from hatch_mypyc.utils import construct_setup_file


class TestConstructSetupFile:
    def test_no_arguments(self):
        assert (
            construct_setup_file(opt_level='3')
            == """\
from setuptools import setup
from mypyc.build import mypycify

setup(
    name='mypyc_output',
    ext_modules=mypycify(
        [
        ],
        opt_level='3',
    ),
)
"""
        )

    def test_one_argument(self):
        assert (
            construct_setup_file('foo', opt_level='3')
            == """\
from setuptools import setup
from mypyc.build import mypycify

setup(
    name='mypyc_output',
    ext_modules=mypycify(
        [
            'foo',
        ],
        opt_level='3',
    ),
)
"""
        )

    def test_multiple_arguments(self):
        assert (
            construct_setup_file('foo', 'bar', opt_level='3')
            == """\
from setuptools import setup
from mypyc.build import mypycify

setup(
    name='mypyc_output',
    ext_modules=mypycify(
        [
            'foo',
            'bar',
        ],
        opt_level='3',
    ),
)
"""
        )

    def test_no_options(self):
        assert (
            construct_setup_file('foo', 'bar')
            == """\
from setuptools import setup
from mypyc.build import mypycify

setup(
    name='mypyc_output',
    ext_modules=mypycify(
        [
            'foo',
            'bar',
        ],
    ),
)
"""
        )



# SPDX-FileCopyrightText: 2021-present Ofek Lev <oss@ofek.dev>
#
# SPDX-License-Identifier: MIT


def construct_setup_file(*args, **kwargs):
    # Ensure that setuptools is imported first
    contents = """\
from setuptools import setup
from mypyc.build import mypycify

setup(
    name='mypyc_output',
    ext_modules=mypycify(
        [{}
        ],{}
    ),
)
"""

    return contents.format(
        ''.join(f'\n            {arg!r},' for arg in args),
        ''.join(f'\n        {key}={value!r},' for key, value in kwargs.items()),
    )



# SPDX-FileCopyrightText: 2021-present Ofek Lev <oss@ofek.dev>
#
# SPDX-License-Identifier: MIT
from hatchling.plugin import hookimpl

from .plugin import MypycBuildHook


@hookimpl
def hatch_register_build_hook():
    return MypycBuildHook



# SPDX-FileCopyrightText: 2021-present Ofek Lev <oss@ofek.dev>
#
# SPDX-License-Identifier: MIT
import zipfile

from packaging.tags import sys_tags

from .utils import build_project


def test_no_exclusion(new_project):
    build_project()

    build_dir = new_project / 'dist'
    assert build_dir.is_dir()

    artifacts = list(build_dir.iterdir())
    assert len(artifacts) == 1
    wheel_file = artifacts[0]

    best_matching_tag = next(sys_tags())
    assert wheel_file.name == f'my_app-1.2.3-{best_matching_tag}.whl'

    extraction_directory = new_project.parent / '_archive'
    extraction_directory.mkdir()

    with zipfile.ZipFile(str(wheel_file), 'r') as zip_archive:
        zip_archive.extractall(str(extraction_directory))

    assert len(list(extraction_directory.iterdir())) == 3

    metadata_directory = extraction_directory / 'my_app-1.2.3.dist-info'
    assert metadata_directory.is_dir()

    wheel_metadata_file = metadata_directory / 'WHEEL'
    assert wheel_metadata_file.is_file()
    assert 'Root-Is-Purelib: false' in wheel_metadata_file.read_text(encoding='utf-8')

    extracted_package_dir = extraction_directory / 'my_app'
    assert extracted_package_dir.is_dir()

    distributed_files = list(extracted_package_dir.iterdir())
    assert len(distributed_files) == 5

    root_files = 0
    fibonacci_files = 0
    for distributed_file in distributed_files:
        if distributed_file.name.startswith('__init__'):
            root_files += 1
        elif distributed_file.name.startswith('fib'):
            fibonacci_files += 1

    assert root_files == 2
    assert fibonacci_files == 2


def test_exclusion(new_project):
    project_file = new_project / 'pyproject.toml'
    contents = project_file.read_text(encoding='utf-8')
    contents += '\nexclude = ["__main__.py"]'
    project_file.write_text(contents, encoding='utf-8')

    package_main = new_project / 'my_app' / '__main__.py'
    package_main.write_text(
        """\
if __name__ == '__main__':
    from .fib import fib
    print(fib())
""",
        encoding='utf-8',
    )

    build_project()

    build_dir = new_project / 'dist'
    assert build_dir.is_dir()

    artifacts = list(build_dir.iterdir())
    assert len(artifacts) == 1
    wheel_file = artifacts[0]

    best_matching_tag = next(sys_tags())
    assert wheel_file.name == f'my_app-1.2.3-{best_matching_tag}.whl'

    extraction_directory = new_project.parent / '_archive'
    extraction_directory.mkdir()

    with zipfile.ZipFile(str(wheel_file), 'r') as zip_archive:
        zip_archive.extractall(str(extraction_directory))

    assert len(list(extraction_directory.iterdir())) == 3

    metadata_directory = extraction_directory / 'my_app-1.2.3.dist-info'
    assert metadata_directory.is_dir()

    wheel_metadata_file = metadata_directory / 'WHEEL'
    assert wheel_metadata_file.is_file()
    assert 'Root-Is-Purelib: false' in wheel_metadata_file.read_text(encoding='utf-8')

    extracted_package_dir = extraction_directory / 'my_app'
    assert extracted_package_dir.is_dir()

    distributed_files = list(extracted_package_dir.iterdir())
    assert len(distributed_files) == 6

    root_files = 0
    fibonacci_files = 0
    for distributed_file in distributed_files:
        if distributed_file.name.startswith('__init__'):
            root_files += 1
        elif distributed_file.name.startswith('fib'):
            fibonacci_files += 1

    assert root_files == 2
    assert fibonacci_files == 2


def test_separation(new_project):
    project_file = new_project / 'pyproject.toml'
    contents = project_file.read_text(encoding='utf-8')
    contents += '\noptions = { separate = true }'
    project_file.write_text(contents, encoding='utf-8')

    build_project()

    build_dir = new_project / 'dist'
    assert build_dir.is_dir()

    artifacts = list(build_dir.iterdir())
    assert len(artifacts) == 1
    wheel_file = artifacts[0]

    best_matching_tag = next(sys_tags())
    assert wheel_file.name == f'my_app-1.2.3-{best_matching_tag}.whl'

    extraction_directory = new_project.parent / '_archive'
    extraction_directory.mkdir()

    with zipfile.ZipFile(str(wheel_file), 'r') as zip_archive:
        zip_archive.extractall(str(extraction_directory))

    assert len(list(extraction_directory.iterdir())) == 2

    metadata_directory = extraction_directory / 'my_app-1.2.3.dist-info'
    assert metadata_directory.is_dir()

    wheel_metadata_file = metadata_directory / 'WHEEL'
    assert wheel_metadata_file.is_file()
    assert 'Root-Is-Purelib: false' in wheel_metadata_file.read_text(encoding='utf-8')

    extracted_package_dir = extraction_directory / 'my_app'
    assert extracted_package_dir.is_dir()

    distributed_files = list(extracted_package_dir.iterdir())
    assert len(distributed_files) == 5

    root_files = 0
    fibonacci_files = 0
    for distributed_file in distributed_files:
        if distributed_file.name.startswith('__init__'):
            root_files += 1
        elif distributed_file.name.startswith('fib'):
            fibonacci_files += 1

    assert root_files == 1
    assert fibonacci_files == 3


def test_target_not_wheel(new_project):
    project_file = new_project / 'pyproject.toml'
    contents = project_file.read_text(encoding='utf-8')
    contents = contents.replace('[tool.hatch.build.targets.wheel.hooks.mypyc]', '[tool.hatch.build.hooks.mypyc]')
    project_file.write_text(contents, encoding='utf-8')

    build_project('-t', 'sdist')

    build_dir = new_project / 'dist'
    assert build_dir.is_dir()

    artifacts = list(build_dir.iterdir())
    assert len(artifacts) == 1

    assert artifacts[0].name.endswith('.tar.gz')



# SPDX-FileCopyrightText: 2021-present Ofek Lev <oss@ofek.dev>
#
# SPDX-License-Identifier: MIT
import os
import platform
from glob import glob

import pytest

from .utils import build_project

pytestmark = [pytest.mark.usefixtures('new_project')]


def test():
    compiled_extension = '.pyd' if platform.system() == 'Windows' else '.so'
    shared_lib_pattern = f'*__mypyc.*{compiled_extension}'
    lib_pattern = f'my_app{os.path.sep}*.*{compiled_extension}'

    assert not glob(shared_lib_pattern)
    assert not glob(lib_pattern)

    build_project()

    assert len(glob(shared_lib_pattern)) == 1
    assert len(glob(lib_pattern)) == 2

    build_project('--clean-only')

    assert not glob(shared_lib_pattern)
    assert not glob(lib_pattern)



# SPDX-FileCopyrightText: 2021-present Ofek Lev <oss@ofek.dev>
#
# SPDX-License-Identifier: MIT
from __future__ import annotations

import os
import platform
import subprocess
import sys
from functools import cached_property
from tempfile import TemporaryDirectory

import pathspec
from hatchling.builders.hooks.plugin.interface import BuildHookInterface

from .utils import construct_setup_file


class MypycBuildHook(BuildHookInterface):
    PLUGIN_NAME = 'mypyc'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._on_windows = platform.system() == 'Windows'

    @cached_property
    def config_mypy_args(self):
        mypy_args = self.config.get('mypy-args', [])
        if isinstance(mypy_args, list):
            for i, argument in enumerate(mypy_args, 1):
                if not isinstance(argument, str):
                    raise TypeError(
                        f'Argument #{i} of option `mypy-args` for build hook `{self.PLUGIN_NAME}` must be a string'
                    )
                elif not argument:
                    raise ValueError(
                        f'Argument #{i} of option `mypy-args` for build hook `{self.PLUGIN_NAME}` '
                        f'cannot be an empty string'
                    )
        else:
            raise TypeError(f'Option `mypy-args` for build hook `{self.PLUGIN_NAME}` must be an array')

        return mypy_args

    @cached_property
    def config_options(self):
        options = self.config.get('options', {})
        if not isinstance(options, dict):
            raise TypeError(f'Option `options` for build hook `{self.PLUGIN_NAME}` must be a table')

        return options

    @cached_property
    def config_include(self):
        patterns = self.config.get('include', [])
        if isinstance(patterns, list):
            if not patterns:
                raise ValueError(f'Option `include` for build hook `{self.PLUGIN_NAME}` is required')

            for i, pattern in enumerate(patterns, 1):
                if not isinstance(pattern, str):
                    raise TypeError(
                        f'Pattern #{i} of option `include` for build hook `{self.PLUGIN_NAME}` must be a string'
                    )
                elif not pattern:
                    raise ValueError(
                        f'Pattern #{i} of option `include` for build hook `{self.PLUGIN_NAME}` '
                        f'cannot be an empty string'
                    )
        else:
            raise TypeError(f'Option `include` for build hook `{self.PLUGIN_NAME}` must be an array')

        return patterns

    @cached_property
    def config_exclude(self):
        patterns = self.config.get('exclude', [])
        if isinstance(patterns, list):
            for i, pattern in enumerate(patterns, 1):
                if not isinstance(pattern, str):
                    raise TypeError(
                        f'Pattern #{i} of option `exclude` for build hook `{self.PLUGIN_NAME}` must be a string'
                    )
                elif not pattern:
                    raise ValueError(
                        f'Pattern #{i} of option `exclude` for build hook `{self.PLUGIN_NAME}` '
                        f'cannot be an empty string'
                    )
        else:
            raise TypeError(f'Option `exclude` for build hook `{self.PLUGIN_NAME}` must be an array')

        return patterns

    @cached_property
    def include_spec(self):
        return pathspec.PathSpec.from_lines(pathspec.patterns.GitWildMatchPattern, self.config_include)

    @cached_property
    def exclude_spec(self):
        if not self.config_exclude:
            return None

        return pathspec.PathSpec.from_lines(pathspec.patterns.GitWildMatchPattern, self.config_exclude)

    def include_path(self, relative_path):
        return self.include_spec.match_file(relative_path) and not self.path_is_excluded(relative_path)

    def path_is_excluded(self, relative_path):
        if self.exclude_spec is None:
            return False

        return self.exclude_spec.match_file(relative_path)

    @cached_property
    def included_files(self):
        included_files = []

        for root, dirs, files in os.walk(self.root):
            relative_path = os.path.relpath(root, self.root)

            # First iteration
            if relative_path == '.':
                relative_path = ''

            dirs[:] = sorted(
                d
                for d in dirs
                # The trailing slash is necessary so e.g. `bar/` matches `foo/bar`
                if not self.path_is_excluded('{}/'.format(os.path.join(relative_path, d)))
            )

            for f in sorted(files):
                if not f.endswith('.py'):
                    continue

                relative_file_path = os.path.join(relative_path, f)
                if self.include_path(relative_file_path):
                    included_files.append(relative_file_path)

        return included_files

    @cached_property
    def normalized_included_files(self):
        if self._on_windows:
            return [f.replace('\\', '/') for f in self.included_files]

        return self.included_files

    @cached_property
    def artifact_globs(self):
        artifact_globs = []
        separation = self.config_options.get('separate', False) is not False
        compiled_extension = '.pyd' if self._on_windows else '.so'

        for included_file in self.included_files:
            root, _ = os.path.splitext(included_file)
            artifact_globs.append(f'{root}.*{compiled_extension}')

            if separation:
                artifact_globs.append(f'{root}__mypyc.*{compiled_extension}')

        if not separation:
            artifact_globs.append(f'*__mypyc.*{compiled_extension}')

        return artifact_globs

    @cached_property
    def normalized_artifact_globs(self):
        if self._on_windows:
            return [f.replace('\\', '/') for f in self.artifact_globs]

        return self.artifact_globs

    @cached_property
    def artifact_patterns(self):
        # Match the exact path starting at the project root
        return [f'/{artifact_glob}' for artifact_glob in self.normalized_artifact_globs]

    def clean(self, versions):
        from glob import iglob

        for artifact_glob in self.artifact_globs:
            absolute_glob = os.path.join(self.root, artifact_glob)
            for artifact in iglob(absolute_glob):
                os.remove(artifact)

    def initialize(self, version, build_data):
        if self.target_name != 'wheel':
            return

        # Hopefully there will be an API for this soon:
        # https://github.com/python/mypy/blob/v0.930/mypyc/__main__.py
        with TemporaryDirectory() as temp_dir:
            temp_dir = os.path.realpath(temp_dir)

            # Prevent temporary files from being written inside the project by both Mypy and setuptools
            shared_temp_build_dir = os.path.join(temp_dir, 'build')
            temp_build_dir = os.path.join(temp_dir, 'tmp')
            os.mkdir(shared_temp_build_dir)
            os.mkdir(temp_build_dir)

            options = self.config_options.copy()
            options['target_dir'] = shared_temp_build_dir

            setup_file = os.path.join(temp_dir, 'setup.py')
            with open(setup_file, 'w', encoding='utf-8') as f:
                f.write(construct_setup_file(*self.config_mypy_args, *self.normalized_included_files, **options))

            process = subprocess.run(
                [
                    sys.executable,
                    setup_file,
                    'build_ext',
                    '--inplace',
                    '--build-lib',
                    shared_temp_build_dir,
                    '--build-temp',
                    temp_build_dir,
                ],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
            )
            if process.returncode:  # no cov
                raise Exception(f'Error while invoking Mypyc:\n{process.stdout.decode("utf-8")}')

        # Success, now finalize build data
        build_data['infer_tag'] = True
        build_data['zip_safe'] = False
        build_data['artifacts'].extend(self.artifact_patterns)
